/* ----- RSTRlintHeap.c ----- */
#//
#// Copyright (c) 1994 - 2017 Norbert Beckmann  <nbinf@kabelmail.de>
#//
#// Licensed under the Apache License, Version 2.0 (see root directory)
#//

/*** heap of Rlint ***/
/* array implementation */


#include <stdlib.h>

#include "RSTRlintHeap.h"
#include "RSTMemAlloc.h"
#include "RSTErrors.h"


/* ----- types ----- */

typedef struct {
               Rint   max;
               Rlint  *arr;
               Rint   qty;
               } RlintHEAPrec;

typedef RlintHEAPrec  *liHEAP;


/* declarations */

static void minLIH_Inst(liHEAP h, Rlint elem, Rint i);
static void maxLIH_Inst(liHEAP h, Rlint elem, Rint i);
static void minLIH_DelMin(liHEAP h, Rlint elem, Rint i);
static void maxLIH_DelMax(liHEAP h, Rlint elem, Rint i);

/************************************************************************/

boolean LIH_New(t_liH *hp,
                Rint max) {

  liHEAP h;
  
  *hp= allocM(sizeof(RlintHEAPrec));
  if (*hp == NULL) {
    pRSTerr("RSTRlintHeap.LIH_New",reAlloc);
    return FALSE;
  }
  h= (liHEAP)*hp;
  
  /* heap starts at arr[1], reserve actually space for max elements: */
  (*h).max= max + 1;
  (*h).qty= 0;
  (*h).arr= allocM((*h).max * sizeof(Rlint));
  if ((*h).arr == NULL) {
    freeM(hp);
    pRSTerr("RSTRlintHeap.LIH_New",reAlloc);
    return FALSE;
  }
  return TRUE;
}

/************************************************************************/

void LIH_Dispose(t_liH *hp) {

  liHEAP h= (liHEAP)*hp;
  
  freeM(&(*h).arr);
  freeM(hp);
}

/************************************************************************/

void LIH_Empty(t_liH hp) {

  (*(liHEAP)hp).qty= 0;
}

/************************************************************************/

Rint LIH_Qty(t_liH hp) {

  return (*(liHEAP)hp).qty;
}

/************************************************************************/

Rint LIH_MaxQty(t_liH hp) {

  return (*(liHEAP)hp).max - 1; /* heap starts at arr[1] */
}

/************************************************************************/

boolean minLIH_Insert(t_liH hp, Rlint elem) {

  liHEAP h= (liHEAP)hp;
  
  (*h).qty++;
  if ((*h).qty == (*h).max) { /* heap starts at arr[1] */
    (*h).qty--;
    return FALSE;
  }
  (*h).arr[0]= elem;
  minLIH_Inst(h,elem,(*h).qty);
  return TRUE;
}

/************************************************************************/

static void minLIH_Inst(liHEAP h, Rlint elem, Rint i) {

  Rint up= i>>1;
  
  if (i == 1 || elem >= (*h).arr[up]) {
    (*h).arr[i]= (*h).arr[0];
  }
  else {
    (*h).arr[i]= (*h).arr[up];
    minLIH_Inst(h,elem,up);
  }
}

/************************************************************************/

boolean minLIH_Pop(t_liH hp, Rlint *elem) {

  liHEAP h= (liHEAP) hp;
  
  if ((*h).qty == 0) {
    return FALSE;
  }
  else {
    *elem= (*h).arr[1];
    minLIH_DelMin(h,(*h).arr[(*h).qty],1);
    (*h).qty--;
    return TRUE;
  }
}

/************************************************************************/

static void minLIH_DelMin(liHEAP h, Rlint elem, Rint i) {

  Rint down= i<<1;
  
  if (down >= (*h).qty || elem <= (*h).arr[down] && elem <= (*h).arr[down+1]) {
    (*h).arr[i]= (*h).arr[(*h).qty];
  }
  else {
    if ((*h).arr[down+1] < (*h).arr[down]) {
      down++;
    }
    (*h).arr[i]= (*h).arr[down];
    minLIH_DelMin(h,elem,down);
  }
}

/************************************************************************/

boolean maxLIH_Insert(t_liH hp, Rlint elem) {

  liHEAP h= (liHEAP)hp;
  
  (*h).qty++;
  if ((*h).qty == (*h).max) { /* heap starts at arr[1] */
    (*h).qty--;
    return FALSE;
  }
  (*h).arr[0]= elem;
  maxLIH_Inst(h,elem,(*h).qty);
  return TRUE;
}

/************************************************************************/

static void maxLIH_Inst(liHEAP h, Rlint elem, Rint i) {

  Rint up= i>>1;
  
  if (i == 1 || elem <= (*h).arr[up]) {
    (*h).arr[i]= (*h).arr[0];
  }
  else {
    (*h).arr[i]= (*h).arr[up];
    maxLIH_Inst(h,elem,up);
  }
}

/************************************************************************/

boolean maxLIH_Pop(t_liH hp, Rlint *elem) {

  liHEAP h= (liHEAP) hp;
  
  if ((*h).qty == 0) {
    return FALSE;
  }
  else {
    *elem= (*h).arr[1];
    maxLIH_DelMax(h,(*h).arr[(*h).qty],1);
    (*h).qty--;
    return TRUE;
  }
}

/************************************************************************/

static void maxLIH_DelMax(liHEAP h, Rlint elem, Rint i) {

  Rint down= i<<1;
  
  if (down >= (*h).qty || elem >= (*h).arr[down] && elem >= (*h).arr[down+1]) {
    (*h).arr[i]= (*h).arr[(*h).qty];
  }
  else {
    if ((*h).arr[down+1] > (*h).arr[down]) {
      down++;
    }
    (*h).arr[i]= (*h).arr[down];
    maxLIH_DelMax(h,elem,down);
  }
}

/************************************************************************/
